#-*- coding: utf-8 -*-

import sys
from collections import Counter

def load_dictionary(fnames):

    ne_dict = []
    label_set = []
    for fname in fnames:
        with open(fname, 'r', encoding='utf-8') as f:
            for l in f:
                tags, words = l.strip().split('\t')
                for i, tag in enumerate(tags.split('/')):
                    while len(label_set) <= i:
                        label_set.append(set())
                        ne_dict.append({})
                    if tag == '':
                        continue
                    if not tag in ne_dict[i]:
                        ne_dict[i][tag] = []
                    ne_dict[i][tag].append(words.split(' '))
    return ne_dict, label_set

def add_ne_dict(ne_dict, words, tags):

    for j in range(len(ne_dict)):
        i = len(words)
        while i>0:
            i -= 1
            if tags[j][i]=='O' or tags[j][i]=='[PAD]':
                continue
            cur_word = []
            cur_tag = tags[j][i][2:] if tags[j][i].startswith('B-') or tags[j][i].startswith('I-') else tags[j][i][:-2]
            while i>=0:
                cur_word.append(words[i])
                if tags[j][i].startswith('B-') or tags[j][i].endswith('-B'):
                    break
                i -= 1
                
            cur_word.reverse()
            if not cur_tag in ne_dict[j]:
                ne_dict[j][cur_tag] = []
            if not cur_word in ne_dict[j][cur_tag]:
                ne_dict[j][cur_tag].append(cur_word)
    

def load_spl(fnames):

    ne_dict = []
    label_set = []
    dataset = []
    for fname in fnames:
        with open(fname, 'r', encoding='utf-8') as f:
            for l in f:
                words_tags = l.strip().split(' ')
                if len(words_tags) == 0 or (len(words_tags) == 1 and words_tags[0] == ''):
                    continue
                n_tagtype = max([len(w.split('/'))-1 for w in words_tags])
                while len(label_set) < n_tagtype:
                    label_set.append(set())
                    ne_dict.append({})
                words = [w.split('/')[0] for w in words_tags]
                tags = [[w.split('/')[i+1] if i+1 < len(w.split('/')) and w.split('/')[i+1] != '' else '[PAD]' for w in words_tags] for i in range(len(label_set))]

                dataset.append((words, tags))
                for i in range(len(tags)):
                    label_set[i] |= set(tags[i])

                # make NE dictionary
                add_ne_dict(ne_dict, words, tags)
                                
    return dataset, ne_dict, label_set

def load_conll(fnames):

    ne_dict = []
    label_set = []
    dataset = []
    for fname in fnames:
        with open(fname, 'r', encoding='utf-8') as f:
            words = []
            tags_rev = []
            for l in f:
                l = l.strip()
                if len(l) == 0:

                    if len(words) == 0:
                        continue

                    n_tagtype = max([len(elem) for elem in tags_rev])
                    while len(label_set) < n_tagtype:
                        label_set.append(set())
                        ne_dict.append({})
                    tags = [[tags_rev[i][j] if len(tags_rev[i]) > j else '[PAD]' for i in range(len(tags_rev))] for j in range(n_tagtype)]
                    dataset.append((words, tags))
                    for i in range(len(tags)):
                        label_set[i] |= set(tags[i])

                    # make NE dictionary
                    add_ne_dict(ne_dict, words, tags)
                    
                    words = []
                    tags_rev = []
                    continue
                l_ary = l.split(' ')
                words.append(l_ary[0])
                tags_rev.append(l_ary[1:])

    return dataset, ne_dict, label_set

def load_dataset(filenames):
    '''
    Args:
        filenames: dict {str -> str}
        mode: str
            spl: a sentence per line (token/tag, delim=' ')
            conll: a token per line (token tag)
    Returns:
        datasets: dict {str -> list of (list, list)}, set(str)
    '''

    file_format = filenames['format']
    datasets = {}
    label_set = []
    ne_dict = []
    for data_type, fnames in filenames.items():
        if data_type == 'format':
            continue
        if len(fnames) == 0:
            continue
        if data_type == 'dict':
            loaded_ne_dict, loaded_label_set = load_dictionary(fnames)
            for i, (nd, ls) in enumerate(zip(loaded_ne_dict, loaded_label_set)):
                while len(label_set) <= i:
                    label_set.append(set())
                    ne_dict.append({})
                label_set[i] |= ls
                for t, wl in nd.items():
                    #print(t)
                    if t == '':
                        continue
                    if not t in ne_dict[i]:
                        ne_dict[i][t] = []
                    for w in wl:
                        if not w in ne_dict[i][t]:
                            ne_dict[i][t].append(w)
            continue
        
        datasets[data_type] = []
        if file_format == 'spl':
            loaded_dataset, loaded_ne_dict, loaded_label_set = load_spl(fnames)
        elif file_format == 'conll':
            loaded_dataset, loaded_ne_dict, loaded_label_set = load_conll(fnames)
        datasets[data_type] += loaded_dataset

        if data_type in ['train']:
            for i, (nd, ls) in enumerate(zip(loaded_ne_dict, loaded_label_set)):
                while len(label_set) <= i:
                    label_set.append(set())
                    ne_dict.append({})
                label_set[i] |= ls
                for t, wl in nd.items():
                    if t == '':
                        continue
                    if not t in ne_dict[i]:
                        ne_dict[i][t] = []
                    for w in wl:
                        if not w in ne_dict[i][t]:
                            ne_dict[i][t].append(w)

    for i in range(len(label_set)):
        label_set_add = set()
        for k in label_set[i]:
            if k.startswith('B-'):
                label_set_add.add('I-'+k[2:])
            elif k.endswith('-B'):
                label_set_add.add(k[:-2]+'-I')
        label_set[i] |= label_set_add
                            
    for data_type, data in datasets.items():
        c = Counter()
        n_sent = 0
        n_word = 0
        n_char = 0
        for d in data:
            words, tags = d
            c.update(words)
            n_sent += 1
            n_word += len(words)
            n_char += sum([len(w) for w in words])

        # Print
        print('### {} ###'.format(data_type))
        print('# sent. = {}'.format(n_sent))
        print('# words = {}'.format(n_word))
        print('# char. = {}'.format(n_char))
        sys.stdout.flush()
                                
    print(label_set)
    return datasets, label_set, ne_dict

